
/**********************************************
 *
 *     ArduinoLogger SD datalogger
 *        Copyright 2011 by Travis Smith
 *        trav@tnhsmith.net
 *        See licensing info below    
 *
 *     Project page:   http://www.tnhsmith.net/travproj/ArduinLogger/
 *
 * Credits:
 *     All libraries (see links below w/ includes) used as published
 *
 * Licensing:
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or 
 *   (at your option) any later version.
 *   
 *   This program is distributed in the hope that it will be useful, but 
 *   WITHOUT ANY WARRANTY; without even the implied warranty of 
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
 *   General Public License for more details.
 *   
 *   You should have received a copy of the GNU General Public License 
 *   along with this program; if not, write to the Free Software Foundation, 
 *   Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *   
 *   http://www.opensource.org/licenses/gpl-2.0.php
 *
 ************************************************/

#define VERSION "0.02"

#include <SdFat.h>             // http://code.google.com/p/sdfatlib/
#include <SdFatUtil.h>   		 		// use functions to print free Ram and strings from flash memory
#include <OneWire.h>           // http://milesburton.com/index.php?title=Dallas_Temperature_Control_Library
#include <DallasTemperature.h> // http://milesburton.com/index.php?title=Dallas_Temperature_Control_Library
#include "RTC.h"

//  ***Customize these:
#define VERBOSE_MESSAGES true  
#define TEMP_LOG_ENABLED true    //Enable temperature logging
//for serial slave:
//#define VERBOSE_MESSAGES false  
//#define TEMP_LOG_ENABLED false    

#define TEMP_LOG_PERIOD_MS 60000  //time between temperature log entries, in mS (if enabled)
#define DEFAULT_FILE_NAME "TempLog.txt"  //8.3 max
#define TEMPERATURE_PRECISION 12  //9-12, 12 takes 750mS!
#define TEMPERATURE_DIGITS 1  //number of sig temperature digits to save/print
#define LED_FLOW_PERIOD_MS 2
#define IN_STR_TERM_CHAR '\n'  //character expected to terminate string inputs
#define OUT_STR_TERM_CHAR 3  //(End Of Text) character to terminate string outputs
#define SERIAL_INPUT_TIMEOUT_MS 500  //mS to wait for each user input char

// pin assignments:
#define ONE_WIRE_PIN 4 
#define SDERR_LED0_PIN 5 //LED0/Red
#define LED1_PIN 6 //LED1/Yel
#define SD_WP_PIN 3 
#define SD_CDN_PIN 2 

// store error strings in flash to save RAM
#define error(s) error_P(PSTR(s))

OneWire oneWire(ONE_WIRE_PIN);
DallasTemperature sensors(&oneWire);
DeviceAddress * pdaDevAddr;  //Pointer to be used for device addresses
unsigned char ucNumTempSensors;  //number of temp sensors

Sd2Card card;
SdVolume volume;
SdFile root;
SdFile file;

char cpFileName[13] = DEFAULT_FILE_NAME;

void setup()  
{
	unsigned char ucSensorCount, ucByteNum;
	
	analogWrite(SDERR_LED0_PIN, 0);   //0-255
	analogWrite(LED1_PIN, 255);   //0-255

	Serial.begin(9600);  
	PgmPrint("Trav's ArduinoLogger Project v");
	PgmPrintln(VERSION);

	SDCardInit(false);

	RTCinit();
	
	// Start up the DallasTemperature library
	sensors.begin();
	ucNumTempSensors = sensors.getDeviceCount();
	PgmPrint("OneWire device(s): ");
	Serial.println(ucNumTempSensors, DEC);
	PgmPrint("Parasite pwr: "); 
	if (sensors.isParasitePowerMode()) PgmPrintln("ON");
	else PgmPrintln("OFF");
		
	//Dynamically allocate memory for device addresses and set to zeros (default)
	pdaDevAddr = (DeviceAddress *)calloc(ucNumTempSensors, sizeof(DeviceAddress)); 
	
	for (ucSensorCount = 0; ucSensorCount < ucNumTempSensors; ucSensorCount++)
	{
		PgmPrint("Dev ");
		Serial.print(ucSensorCount, DEC);
		PgmPrint(" Addr: ");
		if (!sensors.getAddress(pdaDevAddr[ucSensorCount], ucSensorCount)) PgmPrintln("Not Found!"); 
		else 
		{
			for (ucByteNum = 0; ucByteNum < 8; ucByteNum++)
	 	   {
	 	      if (pdaDevAddr[ucSensorCount][ucByteNum] < 16) PgmPrint("0");
		      Serial.print(pdaDevAddr[ucSensorCount][ucByteNum], HEX);
 			   PgmPrint(" ");
		   }
		}
		sensors.setResolution(pdaDevAddr[ucSensorCount], TEMPERATURE_PRECISION);
		PgmPrint(", Res: ");
		Serial.print(sensors.getResolution(pdaDevAddr[ucSensorCount]), DEC); 
		PgmPrintln(" bit");
	}

	//write file header row
	if (TEMP_LOG_ENABLED) 
	{
		if (LogFileInit()) 
		{
			for (ucSensorCount = 0; ucSensorCount < ucNumTempSensors; ucSensorCount++)
			{
				file.print(','); 
				for (ucByteNum = 0; ucByteNum < 8; ucByteNum++)
		 	   {
		 	     if (pdaDevAddr[ucSensorCount][ucByteNum] < 16) file.print('0');
			     file.print(pdaDevAddr[ucSensorCount][ucByteNum], HEX);
			   }
			}
			file.println(",Source");  
			if (file.writeError) error("write failed");
			if (!file.close()) error("close failed");
		}
	}

	LogTimeTemp("Start/Reset"); 
	
	PgmPrint("Setup Free RAM: ");
	Serial.println(FreeRam());  
	PgmPrint("\n>");
}


void loop()                     
{
	unsigned long ulLastLogMillis = millis();
	char cVal, cLastSec = -1;
	stcTime curTime;
	
	while(1)
	{
		//blink LED based on RTC
		GetTime(&curTime);
		if (curTime.tSeconds != cLastSec)
		{
			cLastSec = curTime.tSeconds;
			analogWrite(LED1_PIN, (64*(cLastSec & 1)));   //0-255
		}

		//Periodic log entry
		if (millis() - ulLastLogMillis > TEMP_LOG_PERIOD_MS)
		{
			ulLastLogMillis = millis();
			LogTimeTemp("Periodic");
		}
	
		//Check for serial input
		if (Serial.available()) 
			{
			cVal = Serial.read();
			switch (cVal | 0x20)  //conv to lower case
			{
				
				case 'h': //Help 
				case '?': //Help 
					PgmPrintln("h/?: This help menu");
					PgmPrintln("f: set File Name"); // 8.3, root dir only, default is DEFAULT_FILE_NAME
					PgmPrintln("a: append file with string");
					PgmPrintln("r: read file(arg is last X KBytes, def is 0/all)");
					PgmPrintln("d: Directory");
					PgmPrintln("m: manual temp log add");
			      PgmPrintln("i: init SD Card");
			      PgmPrintln("t: read rtc time");
					PgmPrintln("u: update rtc time");
					PgmPrintln("c: CD/WP bit status ");
					break;
				case 'f': //Set File Name  
					NewFilename();
					break;
				case 'a': //append file with string
					AppendFileFromSerial();
					break;
				case 'r': //read file
					ReadFile();
					break;
				case 'd': //Directory
					PgmPrintln("Files/size in root dir:");
					root.ls(LS_SIZE);
					PgmPrintln("-done-");
					Serial.print((char)OUT_STR_TERM_CHAR);
					break;
				case 'i': //init
					SDCardInit(true);
					break;
				case 'm': //manual temp log string
					LogTimeTemp("Manual");
					break;
		      case 't': //read rtc time
					GetTime(&curTime);
					SerialPrintTime(curTime);
					break;
				case 'u': //update rtc time
					if (SerialInputTime(&curTime))
					{
						PgmPrint("Entered: ");
						SerialPrintTime(curTime);
						SetTime(curTime);
						PgmPrint("Saved, Read: ");
						GetTime(&curTime);  
						SerialPrintTime(curTime);
					}
					else
					{
						PgmPrintln("Bad format, time not updated");
					}
					break;
				case 'c': //CD/WP bit status 
      			PgmPrint("WP: ");
					Serial.print(digitalRead(SD_WP_PIN));
      			PgmPrint("\tCDN: ");
					Serial.println(digitalRead(SD_CDN_PIN));
					delay(200);
					break;
	      }
      PgmPrint(">"); //prompt
		} //serial available
	}  
}


void FilePrintPad(char cNum)
{
  if (cNum < 10) file.print('0'); 
  file.print(cNum, DEC); 
}


void error_P(const char* str) 
{
	PgmPrint("error: ");
	SerialPrint_P(str);
	PgmPrintln("!!!");
	if (card.errorCode()) 
	{
		PgmPrint("SD error: ");
		Serial.print(card.errorCode(), HEX);
		Serial.print(',');
		Serial.println(card.errorData(), HEX);
	}
	digitalWrite(SDERR_LED0_PIN, HIGH);
	//while(1);  //can we recover from an error?
}


void NewFilename()
{
	char cpNewName[13];
	
	if (SerialInputString(cpNewName, 13))
	{
		strcpy(cpFileName, cpNewName);
		PgmPrint("FN set");  
		if (VERBOSE_MESSAGES)
		{
			PgmPrint(" to: ");  
			Serial.print(cpFileName);
		}
		Serial.println();  
	}
}


bool SerialInputString(char * cpInBuf, int iMaxLen)
{
	char cCharnum = 0;

	do
	{
		if (!SerialAvailabeTimeout()) return(false);
		cpInBuf[cCharnum] = Serial.read();
		cCharnum++;
	} while (cCharnum < iMaxLen && cpInBuf[cCharnum - 1] != IN_STR_TERM_CHAR);
	
	if (cpInBuf[cCharnum - 1] == IN_STR_TERM_CHAR)
	{
		cpInBuf[cCharnum - 1] = 0; //terminate string
		return(true);
	}
	else
	{
		PgmPrintln("\nName too long");
		return(false);  
	}	
}


void ReadFile()
{
	int16_t cIn;
	char cpKBtoPrint[5];
	long lKBtoPrint, lFileSizeKB;

	if (SerialInputString(cpKBtoPrint, 5))
	{
		lKBtoPrint = atol(cpKBtoPrint);

		if (!file.open(&root, cpFileName, O_READ)) error("file.open failed");
		else
		{	
			lFileSizeKB = file.fileSize()/1024UL;
			Serial.println(cpFileName);
			Serial.print((int)(lFileSizeKB+1));
			PgmPrintln("KB");
		
			if (lKBtoPrint > 0 && lFileSizeKB - lKBtoPrint > 0)
			{
				if (!file.seekSet((unsigned long)(lFileSizeKB - lKBtoPrint)*1024UL)) error("seekSet fail");
				PgmPrint("Last ");
				Serial.print(lKBtoPrint);
				PgmPrintln("KB:");
			}
			else	
			{
				PgmPrintln("Full File:");
			}
			
			while ((cIn = file.read()) > 0) Serial.print((char)cIn);
			
			PgmPrintln("-done-");	
			Serial.print((char)OUT_STR_TERM_CHAR);
			if (!file.close()) error("close failed");
		}
	}
}


void AppendFileFromSerial()
{
	char cIn;
	bool bNoTermChar = true, bNoSerTO = true;
	stcTime myTime;
	
	file.writeError = false; // clear write error
	
	if (!file.open(&root, cpFileName, O_CREAT | O_APPEND | O_WRITE)) error("open failed");

	while (bNoSerTO && bNoTermChar)
	{
		if (SerialAvailabeTimeout())
		{
			cIn = Serial.read();
			if (cIn == IN_STR_TERM_CHAR) bNoTermChar = false;
			else file.print(cIn);
		}
		else bNoSerTO = false;
	} 

	if (!bNoSerTO) file.print("-TO-");
		
	file.println(); //end line
	
	GetTime(&myTime);  
	// set write/modification date time:
	if (!file.timestamp(T_WRITE, (2000 + myTime.tYear), myTime.tMonth, myTime.tDate, \
		(myTime.tHours == 12 ? (12 * myTime.tAMPM) : (myTime.tHours + 12 * myTime.tAMPM)), \
		myTime.tMinutes, myTime.tSeconds)) error("file time failed");
	
	if (file.writeError) error("write failed");
	if (!file.close()) error("close failed");

	PgmPrintln("Appended");
	if (VERBOSE_MESSAGES) Serial.println(cpFileName);

}


void LogTimeTemp(const char * cpSource)
{
	float fInTempF;
	unsigned char ucSensorNum;
	
	if (!TEMP_LOG_ENABLED) return;
		
	if (!LogFileInit()) return; //early return

	sensors.requestTemperatures();  //this call takes 750mS if in 12 bit sampling mode
	
	for (ucSensorNum = 0; ucSensorNum < ucNumTempSensors; ucSensorNum++)
	{
		file.print(',');  
		fInTempF = DallasTemperature::toFahrenheit(sensors.getTempC(pdaDevAddr[ucSensorNum]));
		file.print(fInTempF, TEMPERATURE_DIGITS);
	}

	file.print(',');  
	file.println(cpSource);
	
	if (file.writeError) error("write failed");
	if (!file.close()) error("close failed");

	if (VERBOSE_MESSAGES)
	{
		stcTime myTime;
		
		Serial.println(cpFileName);
		GetTime(&myTime);  
		SerialPrintTime(myTime);
		for (ucSensorNum = 0; ucSensorNum < ucNumTempSensors; ucSensorNum++)
		{
			fInTempF = DallasTemperature::toFahrenheit(sensors.getTempC(pdaDevAddr[ucSensorNum]));
			Serial.print(fInTempF, TEMPERATURE_DIGITS);
			PgmPrint(", ");  
		}
		Serial.println(cpSource);
		PgmPrint("TLFree RAM: ");
		Serial.println(FreeRam());  
	}
}


boolean LogFileInit()
{
	stcTime myTime;
	char cHours24;
	
	if (digitalRead(SD_CDN_PIN))
	{
		error("No SD Card");
		return(false); //early exit
	}
	
	if (digitalRead(SD_WP_PIN))
	{
		error("Write Prot");
		return(false); //early exit
	}

	// clear write error
	file.writeError = false;
	
	// O_CREAT - create the file if it does not exist
	// O_APPEND - seek to the end of the file prior to each write
	// O_WRITE - open for write
	if (!file.open(&root, cpFileName, O_CREAT | O_APPEND | O_WRITE)) error("open failed");

	GetTime(&myTime);  
	//convert to 24 hr:
	cHours24 = (myTime.tHours == 12 ? (12 * myTime.tAMPM) : (myTime.tHours + 12 * myTime.tAMPM));       
	//set write/modification date time:
	if (!file.timestamp(T_WRITE, (2000 + myTime.tYear), myTime.tMonth, myTime.tDate, cHours24, myTime.tMinutes, myTime.tSeconds)) error("file time failed");
  	
   // date/time in 24 hr format:
	file.print((2000 + myTime.tYear), DEC);       
	file.print('/');    
	FilePrintPad(myTime.tMonth);       
	file.print('/');    
	FilePrintPad(myTime.tDate);  
	file.print(' ');    
	FilePrintPad(cHours24);       
	file.print(':');   
	FilePrintPad(myTime.tMinutes);       
	file.print(':');    
	FilePrintPad(myTime.tSeconds); 
	
	return(true);
}


void SDCardInit(bool bCloseRoot)
{
	unsigned long ulStartMillis = millis();
	
	if (digitalRead(SD_CDN_PIN) || digitalRead(SD_WP_PIN))
	{
		error("No SD/WP");
		return; //early exit
	}
	
	if (bCloseRoot) if (!root.close()) error("root.close failed");
		
	//pinMode(A0, OUTPUT); //debug pin
	//digitalWrite(A0, HIGH); //debug pin 
	// initialize the SD card at SPI_HALF_SPEED to avoid bus errors with
	// breadboards.  use SPI_FULL_SPEED for better performance.
	//These 3 init lines characterization:
	//   no card: 2.644-2.824 Sec
	//   1GB uSD card: 8.17mS  1ea 5M file + other small ones ; first insertion: 147mS, 447mS if re-inserted?
	//   1GB uSD card: 8.17mS 32ea 5M file + other small ones ; first insertion: 140mS
	//   2GB uSD card: 10.01mS first insertion: 365mS
	if (!card.init(SPI_FULL_SPEED)) error("card.init failed");
	// initialize a FAT volume
	if (!volume.init(&card)) error("volume.init failed");
	// open the root directory
	if (!root.openRoot(&volume)) error("openRoot failed");
	//digitalWrite(A0, LOW); //debug pin 
	
	PgmPrint("Init");
	if (VERBOSE_MESSAGES)
	{
		ulStartMillis = millis() - ulStartMillis;
		PgmPrint(" took ");
		Serial.print(ulStartMillis);
		PgmPrint("mS");
	}
	Serial.println();  

}